import copy
import hashlib
import os
import tarfile
import libvirt
import tempfile
import shutil

from virtinst import xmlutil
from virtinst.snapshot import DomainSnapshot
from virtinst.storage import StoragePool, StorageVolume

from .devices.interface import DeviceInterface

from .logger import log
from .guest import Guest


def _upload_disk(vol, st, filename):
    def bytesReadHandler(stream, nbytes, opaque):
        fd = opaque
        return os.read(fd, nbytes)

    def holeHandler(stream, opaque):
        fd = opaque
        cur = os.lseek(fd, 0, os.SEEK_CUR)

        try:
            data = os.lseek(fd, cur, os.SEEK_DATA)
        except OSError as e:
            if e.errno != 6:
                raise e
            else:
                data = -1
        if data < 0:
            inData = False
            eof = os.lseek(fd, 0, os.SEEK_END)
            if (eof < cur):
                raise RuntimeError("Current position in file after EOF: %d" % cur)
            sectionLen = eof - cur
        else:
            if (data > cur):
                inData = False
                sectionLen = data - cur
            else:
                inData = True

                hole = os.lseek(fd, data, os.SEEK_HOLE)
                if hole < 0:
                    raise RuntimeError("No trailing hole")

                if (hole == data):
                    raise RuntimeError("Impossible happened")
                else:
                    sectionLen = hole - data
        os.lseek(fd, cur, os.SEEK_SET)
        return [inData, sectionLen]

    def sendSkipHandler(stream, length, opaque):
        fd = opaque
        return os.lseek(fd, length, os.SEEK_CUR)

    offset = 0
    length = 0

    fd = os.open(filename, os.O_RDONLY)
    vol.upload(st, offset, length, libvirt.VIR_STORAGE_VOL_UPLOAD_SPARSE_STREAM)
    st.sparseSendAll(bytesReadHandler, holeHandler, sendSkipHandler, fd)

    os.close(fd)


class Importer(object):
    def __init__(self, conn):
        self.conn = conn
        self._new_guest = None
        self._tmp = None


    #################
    # Init routines #
    #################

    def _init_new_guest(self, src_xml, pool, save_mac):
        """
        Perform the series of unconditional new VM changes we always make
        """
        self._new_guest = Guest(self.conn, parsexml=src_xml)
        self._new_guest.uuid = None
        self._new_guest.uuid = Guest.generate_uuid(self.conn)

        pool = StoragePool(self.conn, pool.XMLDesc(0))
        for disk in self._new_guest.devices.disk:
            try:
                disk_path, disk_name = os.path.basename(disk.source.file)
                disk.source.file = os.path.join(pool.target_path, disk_name)
            except:
                pass

        for iface in self._new_guest.devices.interface:
            iface.target_dev = None
            if not save_mac:
                iface.macaddr = DeviceInterface.generate_mac(self.conn)


    ######################
    # Functional methods #
    ######################

    def start_importing(self, tar_path, pool_name, checksum, save_mac):
        """
        Actually perform the importing: unarchiving and defining
        the new import xml.
        """
        log.debug("Starting importing.")

        self._tmp = self._unarchive(tar_path)
        vm = None
        disks = []
        loader = None
        try:
            if checksum:
                if self._compare_checksum(tar_path) == False:
                    raise Exception(_("Checksums do not match"))

            pool = self.conn.storagePoolLookupByName(pool_name)
            if pool is None:
                raise

            log.debug("Importing files: %s", os.listdir(self._tmp))
            for file in os.listdir(self._tmp):
                file_name, file_ext = os.path.splitext(file)
                if file_ext == ".xml":
                    vm = self.save_xml(file, pool, save_mac)

                if file_ext in [".qcow2", ".img", ".iso", ".raw"]:
                    self.save_disk(file, pool)
                    disks.append(file)
                
                if file_ext == ".fd":
                    loader = self.save_loader(file)

            self.save_snapshots(vm, pool)

        except Exception as e:
            log.debug("Importing failed: %s", str(e))
            if vm:
                vm.undefine()
            if disks:
                for disk in disks:
                    pool = StoragePool(self.conn, pool.XMLDesc(0))
                    path = os.path.join(pool.target_path, disk) 
                    vol = self.conn.storageVolLookupByPath(path)
                    vol.delete(0)
            if loader:
                os.remove(loader)

            raise

        finally:
            shutil.rmtree(self._tmp)

        log.debug("Importing finished.")

    def _unarchive(self, tar_path):
        try:
            tmp = tempfile.mkdtemp()
            with tarfile.open(tar_path) as tar:
                tar.extractall(path=tmp)
            return tmp

        except Exception as e:
            log.debug("Unarchiving failed: %s", str(e))
            raise

    def _compare_checksum(self, tar_path):
        md5 = hashlib.md5()

        old_checksum = ""
        with open(f"{tar_path}.md5", "r") as f:
            old_checksum = f.read()

        with open(tar_path, "rb") as f:
            while True:
                chunk = f.read(4096)
                if not chunk:
                    break
                md5.update(chunk)

        new_checksum = md5.hexdigest()
        return bool(new_checksum == old_checksum)

    def save_xml(self, file_name, pool, save_mac):
        with open(os.path.join(self._tmp, file_name), "r") as f:
            self._init_new_guest(f.read(), pool, save_mac)

        return self.conn.defineXML(self._new_guest.get_xml())
    
    def save_loader(self, file_name):
        save_path = "/usr/share/OVMF/"
        shutil.copy(os.path.join(self._tmp, file_name), save_path)
        return os.path.join(save_path, file_name)

    def save_disk(self, file_name, pool):
        st = self.conn.newStream()
        cap = .000001

        vol = StorageVolume(self.conn)
        vol.pool = pool
        vol.name = file_name
        vol.capacity = (cap * 1024 * 1024 * 1024)
        vol.allocation = vol.capacity
        vol.install(meter=None)
        vol = pool.storageVolLookupByName(file_name)

        _upload_disk(vol, st, os.path.join(self._tmp, file_name))
    
    def save_snapshots(self, vm, pool):
        snapshots_dir = os.path.join(self._tmp, "snapshots")
        snaps = []
        pool = StoragePool(self.conn, pool.XMLDesc(0))
        for snapfile in os.listdir(snapshots_dir):
            with open(os.path.join(snapshots_dir, snapfile), "r") as f:
                snap = DomainSnapshot(self.conn, parsexml=f.read())
                snap.domain.name = self._new_guest.name
                snap.domain.uuid = None
                snap.domain.uuid = self._new_guest.uuid

                snap.inactiveDomain.name = self._new_guest.name
                snap.inactiveDomain.uuid = None
                snap.inactiveDomain.uuid = self._new_guest.uuid

                for domain_disk, inactive_disk, disk in zip(snap.domain.devices.disk,
                                             snap.inactiveDomain.devices.disk,
                                             self._new_guest.devices.disk):
                    domain_disk.source.file = disk.source.file
                    inactive_disk.source.file = disk.source.file

                for domain_iface, inactive_iface, iface in zip(snap.domain.devices.interface,
                                             snap.inactiveDomain.devices.interface,
                                             self._new_guest.devices.interface):
                    domain_iface.target_dev = None
                    domain_iface.macaddr = iface.macaddr
                    inactive_iface.target_def = None
                    inactive_iface.target_def = iface.macaddr

                snaps.append(snap)

        snaps.sort(key=lambda snap: snap.creationTime)
        for snap in snaps:
            vm.snapshotCreateXML(snap.get_xml(),
                                 libvirt.VIR_DOMAIN_SNAPSHOT_CREATE_REDEFINE)
