import argparse
import os
import re
import sys

from . import cli
from .cli import fail, print_stdout, print_stderr
from .exporter import Exporter


class SnapshotAdapter:
    def __init__(self, snap):
        self.snap = snap
        self._xmlobj = None

    def get_name(self):
        return self.snap.getName()
    
    def get_xml(self):
        return self.snap.getXMLDesc()

    @property
    def xmlobj(self):
        return SnapshotAdapter(self.snap)


def _process_src(options, conn):
    src_name = options.src_name
    src_xml = None
    domain = None
    if options.original_xml:
        src_xml = open(options.original_xml).read()
        src_name = re.search(r'<name>(.*?)</name>', src_xml).group(1)
    if not src_name:
        fail(_("An original machine name is required,"
            " use '--original src_name' and try again."))
    domain, inactive_xmlobj, active_xmlobj = cli.get_domain_and_guest(
            conn, src_name)
    if not src_xml:
        src_xml = inactive_xmlobj.get_xml()
    return src_name, src_xml, domain


def _process_start_args(conn, domain, options):
    name = options.new_name
    path = options.export_path
    targz_name = os.path.join(path, f"{name}.tar.gz")
    snapshots = []
    for rawsnap in domain.listAllSnapshots():
        snap = SnapshotAdapter(rawsnap)
        snapshots.append(snap)

    return targz_name, snapshots


def parse_args():
    desc = _("Export a virtual machine, changing all the unique "
        "host side configuration like MAC address, name, etc. \n\n"
        "The VM contents are NOT altered: virt-export does not change "
        "anything _inside_ the guest OS, it only downloads disks and "
        "does host side changes. So things like changing passwords, "
        "changing static IP address, etc are outside the scope of "
        "this tool. For these types of changes, please see virt-sysprep(1).")
    parser = cli.setupParser("%(prog)s --connect NAME --original NAME "
                             "--path PATH ...", desc)
    cli.add_connect_option(parser)

    geng = parser.add_argument_group(_("General Options"))
    geng.add_argument("-o", "--original", dest="src_name",
                    help=_("Name of the original guest to export."))
    geng.add_argument("--original-xml",
                    help=_("XML file to use as the original guest."))
    geng.add_argument("-p", "--path", dest="export_path", required=True,
                      help=_("Path for saving archive"))
    geng.add_argument("-n", "--name", dest="new_name", required=True,
                    help=_("Name for the new guest"))
    geng.add_argument("-u", "--uuid", dest="new_uuid",
                      help=_("UUID for the guest"))
    geng.add_argument("--reflink", action="store_true",
            help=_("use btrfs COW lightweight copy"))
    geng.add_argument("--compression", type=int, choices=range(0, 10),
                      default=1, dest="compress",
                      help=_("Choose compression in range 0-9"))
    geng.add_argument("-c", "--checksum", action="store_true",
                      help=_("Create a checksum file"))

    stog = parser.add_argument_group(_("Storage Configuration"))
    stog.add_argument("--force-copy", dest="target", action="append",
                    help=_("Force to copy devices (eg, if 'hdc' is a "
                           "readonly cdrom device, --force-copy=hdc)"))
    stog.add_argument("--skip-copy", action="append",
                    help=_("Skip copy of the device target. (eg, if 'vda' is a "
                           "disk you don't want to copy and use the same path "
                           "in the new VM, use --skip-copy=vda)"))
    stog.add_argument("--nonsparse", action="store_false", dest="sparse",
                    default=True,
                    help=_("Do not use a sparse file for the export's "
                           "disk image"))
    stog.add_argument("--preserve-data", dest="preserve",
            action="store_true", default=False,
            help=_("Do not export storage contents to specified file paths, "
                   "their contents will be left untouched. "
                   "This requires specifying existing paths for "
                   "every exportable disk image."))
    stog.add_argument("--nvram", dest="new_nvram",
                      help=_("New file to use as storage for nvram VARS"))
    

    misc = parser.add_argument_group(_("Miscellaneous Options"))

    # Just used for export tests
    misc.add_argument("--__test-nodry", action="store_true", dest="test_nodry",
                      default=False, help=argparse.SUPPRESS)

    cli.add_misc_options(misc, prompt=True, replace=False, printxml=False)

    cli.autocomplete(parser)

    return parser.parse_args()


def main(conn=None):
    cli.earlyLogging()
    options = parse_args()

    options.quiet = options.quiet
    cli.setupLogging("virt-export", options.debug, options.quiet)

    cli.convert_old_force(options)
    cli.parse_check(options.check)
    cli.set_prompt(options.prompt)
    conn = cli.getConnection(options.connect, conn=conn)

    src_name, src_xml, domain = _process_src(options, conn)
    exporter = Exporter(conn, src_name, src_xml)

    exporter.set_reflink(bool(options.reflink))
    exporter.set_sparse(bool(options.sparse))

    if options.new_uuid is not None:
        exporter.set_export_uuid(options.new_uuid)
    if options.new_nvram:
        exporter.set_nvram_path(options.new_nvram)

    force_targets = options.target or []
    skip_targets = options.skip_copy or []
    for diskinfo in exporter.get_diskinfos():
        if diskinfo.disk.target in force_targets:
            diskinfo.set_clone_requested()
        if diskinfo.disk.target in skip_targets:
            diskinfo.set_share_requested()

    if options.preserve:
        for diskinfo in exporter.get_nonshare_diskinfos():
            diskinfo.set_preserve_requested()
        if exporter.nvram_diskinfo:
            exporter.nvram_diskinfo.set_preserve_requested()

    if options.new_name:
        exporter.set_export_name(options.new_name)

    exporter.prepare()

    run = True
    if run:
        targz_name, snapshots = _process_start_args(conn, domain, options)
        exporter.start_exporting(targz_name, snapshots, options.compress,
                                 bool(options.checksum), cli.get_meter())
        print_stdout("")
        print_stdout(_("Export '%s' created successfully.")
                     % exporter.new_guest.name)

    return 0


def runcli():  # pragma: no cover
    try:
        sys.exit(main())
    except SystemExit as sys_e:
        sys.exit(sys_e.code)
    except KeyboardInterrupt:
        print_stderr(_("Installation aborted at user request"))
    except Exception as main_e:
        fail(main_e)
